#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <stdio.h>
#include <poll.h>

#include <postgres.h>
#include <postgres_fe.h>
#include <libpq-fe.h>

const char *CONNECTION_STR = "postgres://postgres:postgres@172.30.4.76/postgres?sslmode=require";
const int PARAM_COUNT = 11;
const int INSERT_ROW_COUNT = 100;
const int MAX_SEQ = 10000;
const int VAL_BUF_SZ = 32;
const int DEFAULT_INSERT_LOOP_COUNT = 100;
const int DEFAULT_UPDATE_LOOP_COUNT = 100;

struct drand48_data drand48_data_buffer = { 0 };

void strCat(char **tmpSql, size_t *sz, char *buf);
void addRowParams(int row, char **tmpSql, size_t *sz);
char *createInsertQuery(int rowNumber);
int createInsertParams(char **paramValues, int *seq);
int createUpdateParams(char **paramValues, int seq);
void handleNotifications(PGconn *pgConn);
bool isResponseOk(PGresult *result);
bool clearResults(PGconn *pgConn);
bool syncPipeline(PGconn *pgConn);
void printPgError(PGconn *pgConn, PGresult *pgRs, int lineNumber);

/* Memory-size aware string concatenation */
void strCat(char **tmpSql, size_t *sz, char *buf)
{
    size_t lenDst = strlen(*tmpSql);
    size_t lenSrc = strlen(buf);

    if ((lenDst + lenSrc) >= *sz)
    {
        *sz = ((lenDst + lenSrc + BUFSIZ - 1) / BUFSIZ) * BUFSIZ;
        *tmpSql = realloc(*tmpSql, *sz);
    }

    strcat(*tmpSql, buf);
}

void addRowParams(int row, char **tmpSql, size_t *sz)
{
    int col = 0;
    int paramNumber = row * PARAM_COUNT + col + 1;
    char buf[32];

    sprintf(buf, "($%d", paramNumber);
    strCat(tmpSql, sz, buf);

    for (col = 1; col < PARAM_COUNT; col++)
    {
        paramNumber = row * PARAM_COUNT + col + 1;
        sprintf(buf, ",$%d", paramNumber);
        strCat(tmpSql, sz, buf);
    }

    strCat(tmpSql, sz, ")");
}

/**
 * Create an insert query such that following to register INSERT_ROW_COUNT rows.
 * 
 * INSERT INTO test_data (seq, data01, data02, data03, data04, data05, data06, data07, data08, data09, data10) VALUES 
 * ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11), ($12, $13, ...), ...
 */
char *createInsertQuery(int rowNumber)
{
    int row;
    size_t sz = BUFSIZ;
    char *tmpSql = malloc(sz);
    strcpy(tmpSql, "INSERT INTO test_data (seq, data01, data02, data03, data04, data05, "
                    "data06, data07, data08, data09, data10) VALUES ");

    for (row = 0; row < rowNumber - 1; row++)
    {
        addRowParams(row, &tmpSql, &sz);
        strCat(&tmpSql, &sz, ",");
    }
    addRowParams(row, &tmpSql, &sz);

    return tmpSql;
}

/* Set parameters for an insert query for INSERT_ROW_COUNT rows */
int createInsertParams(char **paramValues, int *seq)
{
    int idx;
    int col;
    int rowSeq = *seq;
    double val;
    char *buf;

    for (idx = 0; idx < INSERT_ROW_COUNT; idx++)
    {
        buf = paramValues[idx * PARAM_COUNT];
        snprintf(buf, VAL_BUF_SZ - 1, "%d", rowSeq++);

        for (col = 1; col < PARAM_COUNT; col++)
        {
            buf = paramValues[idx * PARAM_COUNT + col];
            drand48_r(&drand48_data_buffer, &val);
            val *= 100.0;
            snprintf(buf, VAL_BUF_SZ - 1, "%.15f", val);
        }
    }

    *seq = rowSeq;
    return INSERT_ROW_COUNT * PARAM_COUNT;
}

/* Set parameters for update query */
int createUpdateParams(char **paramValues, int seq)
{
    int col;
    double val;
    char *buf;

    buf = paramValues[0];
    snprintf(buf, VAL_BUF_SZ - 1, "%d", seq);

    for (col = 1; col < PARAM_COUNT; col++)
    {
        buf = paramValues[col];
        drand48_r(&drand48_data_buffer, &val);
        val *= 100.0;
        snprintf(buf, VAL_BUF_SZ - 1, "%.15f", val);
    }

    return PARAM_COUNT;
}

void handleNotifications(PGconn *pgConn)
{
    PGnotify *notify;

    /* Consume all notifications, even when there is no function registered */
	while ((notify = PQnotifies(pgConn)) != NULL)
    {
        PQfreemem(notify);
    }
}

bool isResponseOk(PGresult *result)
{
	ExecStatusType resultStatus = PQresultStatus(result);

	bool ok =
		resultStatus == PGRES_SINGLE_TUPLE ||
		resultStatus == PGRES_TUPLES_OK ||
		resultStatus == PGRES_COPY_BOTH ||
		resultStatus == PGRES_COMMAND_OK;

	return ok;
}

/*
 * Consumes results on a connection until NULL is returned.
 * If an error is returned it returns false.
 */
bool clearResults(PGconn *pgConn)
{
	if (PQstatus(pgConn) == CONNECTION_BAD)
	{
		PQfinish(pgConn);
		return false;
	}

	while (true)
	{
		/*
		 * Per Postgres documentation: You should, however, remember to check
		 * PQnotifies after each PQgetResult or PQexec, to see if any
		 * notifications came in during the processing of the command.
		 */
		(void) handleNotifications(pgConn);

		PGresult *result = PQgetResult(pgConn);

		if (result == NULL)
		{
			/* one last time */
			(void) handleNotifications(pgConn);
			break;
		}

		if (!isResponseOk(result))
		{
			PQclear(result);
			PQfinish(pgConn);
			return false;
		}

		PQclear(result);
	}

	return true;
}

bool syncPipeline(PGconn *pgConn)
{
	if (PQpipelineSync(pgConn) != 1)
	{
        printPgError(pgConn, NULL, __LINE__);
		return false;
	}

	int sock = PQsocket(pgConn);
    if (sock < 0)
    {
        printPgError(pgConn, NULL, __LINE__);
        return false;
    }

	bool syncReceived = false;

	/*
	 * PQpipelineSync() might clear the select read readiness, so we need to
	 * consume the input on the first iteration.
	 *
	 * The next iterations will consume input only when poll() returns read
	 * readiness.
	 */
	bool readyToConsume = true;

	int results = 0;

    while (!syncReceived)
    {
		/*
		 * PQisBusy will not itself attempt to read data from the server;
		 * therefore PQconsumeInput must be invoked first, or the busy state
		 * will never end.
		 * Reference: https://www.postgresql.org/docs/current/libpq-async.html
		 */
		if (readyToConsume || PQisBusy(pgConn) == 1)
		{
			if (PQconsumeInput(pgConn) == 0)
			{
                printPgError(pgConn, NULL, __LINE__);
				return false;
			}

			/*
			 * On pipeline mode, we are not worried about the order of
			 * the notifications, we just want to consume them to avoid
			 * filling the notification buffer.
			 */
			(void) handleNotifications(pgConn);
		}

		/*
		 * Read results when the command is not busy.
		 */
		while (PQisBusy(pgConn) == 0)
		{
			PGresult *res = PQgetResult(pgConn);

            if (res == NULL)
            {
				/*
				 * NULL represents a end of result for a single query, but we are
				 * in pipeline mode and we can have multiple results.
				 *
				 * We need to continue consuming until we get a SYNC message.
				 */

				continue;
            }

			ExecStatusType resultStatus = PQresultStatus(res);

			if (resultStatus == PGRES_PIPELINE_SYNC)
			{
				syncReceived = true;

				PQclear(res);

                break;
			}

			if (!isResponseOk(res))
			{
                printPgError(pgConn, res, __LINE__);
				return false;
			}

			PQclear(res);
        }

		/*
		 * We need to wait for the socket to be ready for reading, otherwise
		 * poll() will return immediately and we will busy loop.
		 */
        struct pollfd pollfd = { 0 };
        /* sleep for 10ms to wait for input on the Postgres socket */
        int timeout = 10;

        pollfd.fd = sock;
        pollfd.events = POLLIN;

		int r = poll(&pollfd, 1, timeout);

		if (r == 0 || (r < 0 && errno == EINTR))
		{
			/* got a timeout or signal. The caller will get back later. */
			readyToConsume = false;

			continue;
		}
		else if (r < 0)
		{
            fprintf(stderr, "Failed to poll a postgresql socket. errno : %d\n", errno);
			if (clearResults(pgConn))
            {
			    PQfinish(pgConn);
            }

			return false;
		}

		/* Else there is actually data on the socket */
		readyToConsume = true;
    }

    return true;
}

void printPgError(PGconn *pgConn, PGresult *pgRs, int lineNumber)
{
	char *sqlstate = NULL;

    fprintf(stderr, "Line : %d\n", lineNumber);

    if (pgRs != NULL)
	{
		sqlstate = PQresultErrorField(pgRs, PG_DIAG_SQLSTATE);
        if (sqlstate != NULL)
        {
            fprintf(stderr, "SQLSTATE : %s\n", sqlstate);
        }
	}

	char *message = PQerrorMessage(pgConn);
    if (message != NULL)
    {
        fprintf(stderr, "%s\n", message);
    }

	if (pgRs != NULL)
	{
		PQclear(pgRs);
	}

    if (clearResults(pgConn))
    {
        PQfinish(pgConn);
    }
}

/**
 * Usage: query-data-send-error [-i insert-loop-count] [-u update-loop-count] [-c connection-string]
 */
int main(int argc, char **argv)
{
    int returnStatus = 1;
    int insertLoopCount = DEFAULT_INSERT_LOOP_COUNT;
    int updateLoopCount = DEFAULT_UPDATE_LOOP_COUNT;
    char *connectionString = NULL;

    int sentQuery = 0;
    int idx;
    int seq = 0;
    int maxSeq = 0;
    int paramCount = 0;
    Oid *paramTypes = NULL;
    char *paramValuesBuffer = (char *)malloc(INSERT_ROW_COUNT * PARAM_COUNT * VAL_BUF_SZ);
    char **paramValues = (char **)malloc(INSERT_ROW_COUNT * PARAM_COUNT * sizeof(char *));

    PGconn *pgConn = NULL;
    PGresult *pgRs = NULL;
    ExecStatusType st;

    struct queryName {
        char *name;
        char *sql;
    } queryList[] = {
        {
            "query1",
            createInsertQuery(INSERT_ROW_COUNT)
        },
        {
            "query2",
            "UPDATE test_data SET data01=$2, data02=$3, data03=$4, data04=$5, data05=$6, "
                    "data06=$7, data07=$8, data08=$9, data09=$10, data10=$11 WHERE seq = $1"
        }
    };

    /* Read command arguments */
    int opt;
    while ((opt = getopt(argc, argv, "i:u:c:")) != -1)
    {
        switch (opt)
        {
        case 'i':
            insertLoopCount = atoi(optarg);
            break;
        case 'u':
            updateLoopCount = atoi(optarg);
            break;
        case 'c':
            connectionString = (char *)calloc(strlen(optarg) + 1, sizeof(char));
            strcpy(connectionString, optarg);
            break;
        }
    }

    if (insertLoopCount <= 0)
    {
        insertLoopCount = DEFAULT_INSERT_LOOP_COUNT;
    }
    if (updateLoopCount <= 0)
    {
        updateLoopCount = DEFAULT_UPDATE_LOOP_COUNT;
    }

    pgConn = PQconnectdb(connectionString == NULL ? CONNECTION_STR : connectionString);

    if (PQstatus(pgConn) != CONNECTION_OK)
    {
        printPgError(pgConn, NULL, __LINE__);
        goto ERR_LBL;
    }

    /* Create a table for register test data */
    pgRs = PQexec(pgConn,
                    "CREATE TABLE IF NOT EXISTS test_data ("
                    "seq integer PRIMARY KEY, "
                    "data01 double precision, "
                    "data02 double precision, "
                    "data03 double precision, "
                    "data04 double precision, "
                    "data05 double precision, "
                    "data06 double precision, "
                    "data07 double precision, "
                    "data08 double precision, "
                    "data09 double precision, "
                    "data10 double precision)");
    st = PQresultStatus(pgRs);

    if (st != PGRES_COMMAND_OK)
    {
        printPgError(pgConn, pgRs, __LINE__);
        goto ERR_LBL;
    }
    PQclear(pgRs);

    /* First delete all data in the table */
    pgRs = PQexec(pgConn, "TRUNCATE TABLE test_data");
    st = PQresultStatus(pgRs);

    if (st != PGRES_COMMAND_OK)
    {
        printPgError(pgConn, pgRs, __LINE__);
        goto ERR_LBL;
    }
    PQclear(pgRs);

    /* Enter pipeline mode */
    if (PQenterPipelineMode(pgConn) != 1)
    {
        printPgError(pgConn, NULL, __LINE__);
        goto ERR_LBL;
    }

    /* Set non-blocking mode */
    if (PQsetnonblocking(pgConn, 1) != 0)
    {
        printPgError(pgConn, NULL, __LINE__);
        goto ERR_LBL;
    }

    /* Setup parameter array for query */
    for (idx = 0; idx < (INSERT_ROW_COUNT * PARAM_COUNT); idx++)
    {
        paramValues[idx] = paramValuesBuffer + idx * VAL_BUF_SZ;
    }

    /* Prepare queries */
    for (idx = 0; idx < 2; idx++)
    {
        if (PQsendPrepare(pgConn, queryList[idx].name, queryList[idx].sql, 0, NULL) == 0)
        {
            printPgError(pgConn, NULL, __LINE__);
            goto ERR_LBL;
        }
    }

    srand48_r(time(NULL), &drand48_data_buffer);

    /* Insert test data */
    for (idx = 0; idx < insertLoopCount; idx++)
    {
        sentQuery = PQsendQueryParams(pgConn, "BEGIN", 0, NULL, NULL, NULL, NULL, 0);
        if (sentQuery == 0)
        {
            printPgError(pgConn, NULL, __LINE__);
            goto ERR_LBL;
        }

        paramCount = createInsertParams(paramValues, &maxSeq);
        if (PQsendQueryPrepared(pgConn, queryList[0].name, paramCount, (const char**)paramValues, NULL, NULL, 0) == 0)
        {
            printPgError(pgConn, NULL, __LINE__);
            goto ERR_LBL;
        }

        sentQuery = PQsendQueryParams(pgConn, "COMMIT", 0, NULL, NULL, NULL, NULL, 0);
        if (sentQuery == 0)
        {
            printPgError(pgConn, NULL, __LINE__);
            goto ERR_LBL;
        }

        if (!syncPipeline(pgConn))
        {
            fprintf(stderr, "@%d\n", __LINE__);
            goto ERR_LBL;
        }
    }

    /* Now we iterate through and update all the records in the table */
    for (idx = 0; idx < updateLoopCount; idx++)
    {
        sentQuery = PQsendQueryParams(pgConn, "BEGIN", 0, NULL, NULL, NULL, NULL, 0);
        if (sentQuery == 0)
        {
            printPgError(pgConn, NULL, __LINE__);
            goto ERR_LBL;
        }

        for (seq = 0; seq < maxSeq; seq++)
        {
            paramCount = createUpdateParams(paramValues, seq);
            if (PQsendQueryPrepared(pgConn, queryList[1].name, paramCount, (const char**)paramValues, NULL, NULL, 0) == 0)
            {
                printPgError(pgConn, NULL, __LINE__);
                goto ERR_LBL;
            }
        }

        sentQuery = PQsendQueryParams(pgConn, "COMMIT", 0, NULL, NULL, NULL, NULL, 0);
        if (sentQuery == 0)
        {
            printPgError(pgConn, NULL, __LINE__);
            goto ERR_LBL;
        }

        if (!syncPipeline(pgConn))
        {
            fprintf(stderr, "@%d\n", __LINE__);
            goto ERR_LBL;
        }
    }

    PQfinish(pgConn);
    returnStatus = 0;

ERR_LBL:
    if (connectionString != NULL)
    {
        free(connectionString);
    }
    free(paramValues);
    free(paramValuesBuffer);
    free(queryList[0].sql);

    return returnStatus;
}
